@@ -1,6 +1,6 @@
 #!/usr/bin/env python
 #
-# Copyright (c) 2010-2020 Marcel Martin <marcel.martin@scilifelab.se>
+# Copyright (c) 2010-2021 Marcel Martin <marcel.martin@scilifelab.se>
 #
 # Permission is hereby granted, free of charge, to any person obtaining a copy
 # of this software and associated documentation files (the "Software"), to deal
@@ -23,7 +23,7 @@
 """
 cutadapt version {version}
 
-Copyright (C) 2010-2020 Marcel Martin <marcel.martin@scilifelab.se>
+Copyright (C) 2010-2021 Marcel Martin <marcel.martin@scilifelab.se>
 
 cutadapt removes adapter sequences from high-throughput sequencing reads.
 
@@ -52,19 +52,23 @@ http://dx.doi.org/10.14806/ej.17.1.200
 Run "cutadapt --help" to see all command-line options.
 See https://cutadapt.readthedocs.io/ for full documentation.
 """
-
+import copy
 import sys
 import time
+import shutil
 import logging
 import platform
 import itertools
-from typing import Tuple, Optional, Sequence, List, Any, Iterator, Union
+import multiprocessing
+from typing import Tuple, Optional, Sequence, List, Any, Iterator, Union, Dict
 from argparse import ArgumentParser, SUPPRESS, HelpFormatter
 
 import dnaio
+import xopen as xopen
 
 from cutadapt import __version__
 from cutadapt.adapters import warn_duplicate_adapters, Adapter, InvalidCharacter
+from cutadapt.json import OneLine, dumps as json_dumps
 from cutadapt.parser import AdapterParser
 from cutadapt.modifiers import (SingleEndModifier, LengthTagModifier, SuffixRemover,
     PrefixSuffixAdder,
@@ -88,15 +92,11 @@ class CutadaptArgumentParser(ArgumentParser):
     """
     class CustomUsageHelpFormatter(HelpFormatter):
         def __init__(self, *args, **kwargs):
-            try:
-                import shutil
-                kwargs['width'] = min(24 + 80, shutil.get_terminal_size().columns)
-            except ImportError:
-                pass
+            kwargs['width'] = min(24 + 80, shutil.get_terminal_size().columns)
             super().__init__(*args, **kwargs)
 
         def add_usage(self, usage, actions, groups, prefix=None):
-            if usage is not SUPPRESS:
+            if usage is not SUPPRESS:  # pragma: no cover
                 args = usage, actions, groups, ''
                 self._add_item(self._format_usage, args)
 
@@ -112,7 +112,7 @@ class CutadaptArgumentParser(ArgumentParser):
         """
         print('Run "cutadapt --help" to see command-line options.', file=sys.stderr)
         print('See https://cutadapt.readthedocs.io/ for full documentation.', file=sys.stderr)
-        self.exit(2, "\n{prog}: error: {message}\n".format(prog=self.prog, message=message))
+        self.exit(2, f"\n{self.prog}: error: {message}\n")
 
 
 class CommandLineError(Exception):
@@ -147,11 +147,10 @@ def get_argument_parser() -> ArgumentParser:
 
     group = parser.add_argument_group("Finding adapters",
         description="Parameters -a, -g, -b specify adapters to be removed from "
-            "each read (or from the first read in a pair if data is paired). "
+            "each read (or from R1 if data is paired-end. "
             "If specified multiple times, only the best matching adapter is "
-            "trimmed (but see the --times option). When the special notation "
-            "'file:FILE' is used, adapter sequences are read from the given "
-            "FASTA file.")
+            "trimmed (but see the --times option). Use notation "
+            "'file:FILE' to read adapter sequences from a FASTA file.")
     group.add_argument("-a", "--adapter", type=lambda x: ("back", x), action="append",
         default=[], metavar="ADAPTER", dest="adapters",
         help="Sequence of an adapter ligated to the 3' end (paired data: of the "
@@ -282,6 +281,8 @@ def get_argument_parser() -> ArgumentParser:
         help="Print only error messages.")
     group.add_argument("--report", choices=('full', 'minimal'), default=None,
         help="Which type of report to print: 'full' or 'minimal'. Default: full")
+    group.add_argument("--json", metavar="FILE",
+        help="Dump report in JSON format to FILE")
     group.add_argument("-o", "--output", metavar="FILE",
         help="Write trimmed reads to FILE. FASTQ or FASTA format is chosen "
             "depending on input. Summary report is sent to standard output. "
@@ -311,27 +312,29 @@ def get_argument_parser() -> ArgumentParser:
             "output to same file as trimmed reads")
 
     group = parser.add_argument_group("Paired-end options", description="The "
-        "-A/-G/-B/-U options work like their -a/-b/-g/-u counterparts, but "
-        "are applied to the second read in each pair.")
+        "-A/-G/-B/-U/-Q options work like their lowercase counterparts, but "
+        "are applied to R2 (second read in pair)")
     group.add_argument("-A", type=lambda x: ("back", x), dest='adapters2',
         action='append', default=[], metavar='ADAPTER',
-        help="3' adapter to be removed from second read in a pair.")
+        help="3' adapter to be removed from R2")
     group.add_argument("-G", type=lambda x: ("front", x), dest='adapters2',
         action='append', default=[], metavar='ADAPTER',
-        help="5' adapter to be removed from second read in a pair.")
+        help="5' adapter to be removed from R2")
     group.add_argument("-B", type=lambda x: ("anywhere", x), dest='adapters2',
         action='append', default=[], metavar='ADAPTER',
-        help="5'/3 adapter to be removed from second read in a pair.")
+        help="5'/3 adapter to be removed from R2")
     group.add_argument("-U", dest='cut2', action='append', default=[], type=int, metavar="LENGTH",
-        help="Remove LENGTH bases from second read in a pair.")
+        help="Remove LENGTH bases from R2")
+    group.add_argument("-Q", dest="quality_cutoff2", default=None, metavar="[5'CUTOFF,]3'CUTOFF",
+        help="Quality-trimming cutoff for R2. Default: same as for R1")
     group.add_argument("-p", "--paired-output", metavar="FILE",
-        help="Write second read in a pair to FILE.")
+        help="Write R2 to FILE.")
     group.add_argument("--pair-adapters", action="store_true",
         help="Treat adapters given with -a/-A etc. as pairs. Either both "
              "or none are removed from each read pair.")
     # Setting the default for pair_filter to None allows us to find out whether
     # the option was used at all.
-    group.add_argument("--pair-filter", metavar='(any|both|first)', default=None,
+    group.add_argument("--pair-filter", default=None,
         choices=("any", "both", "first"),
         help="Which of the reads in a paired-end read have to match the "
             "filtering criterion in order for the pair to be filtered. "
@@ -365,7 +368,7 @@ def parse_cutoffs(s: str) -> Tuple[int, int]:
     try:
         cutoffs = [int(value) for value in s.split(",")]
     except ValueError as e:
-        raise CommandLineError("Quality cutoff value not recognized: {}".format(e))
+        raise CommandLineError(f"Quality cutoff value not recognized: {e}")
 
     if len(cutoffs) == 1:
         cutoffs = [0, cutoffs[0]]
@@ -394,9 +397,9 @@ def parse_lengths(s: str) -> Tuple[Optional[int], ...]:
     try:
         values = tuple(int(f) if f != '' else None for f in fields)
     except ValueError as e:
-        raise CommandLineError("Value not recognized: {}".format(e))
+        raise CommandLineError(f"Value not recognized: {e}")
     if len(values) == 2 and values[0] is None and values[1] is None:
-        raise CommandLineError("Cannot parse {!r}: At least one length needs to be given".format(s))
+        raise CommandLineError(f"Cannot parse '{s}': At least one length needs to be given")
     return tuple(values)
 
 
@@ -430,7 +433,11 @@ def open_output_files(
     demultiplex_mode = determine_demultiplex_mode(args)
     if demultiplex_mode and args.discard_trimmed:
         raise CommandLineError("Do not use --discard-trimmed when demultiplexing.")
-
+    if demultiplex_mode == "combinatorial" and args.pair_adapters:
+        raise CommandLineError(
+            "With --pair-adapters, you can only use {name} in your output file name template, "
+            "not {name1} and {name2} (no combinatorial demultiplexing)."
+        )
     if demultiplex_mode == "normal":
         out = out2 = None
         combinatorial_out = combinatorial_out2 = None
@@ -566,7 +573,9 @@ def determine_paired(args) -> bool:
         or args.pair_filter
         or args.untrimmed_paired_output
         or args.too_short_paired_output
-        or args.too_long_paired_output)
+        or args.too_long_paired_output
+        or args.quality_cutoff2
+    )
 
 
 def setup_input_files(
@@ -583,7 +592,7 @@ def setup_input_files(
             "You provided {} input file names, but either one or two are expected. ".format(
                 len(inputs))
             + "The file names were:\n - "
-            + "\n - ".join("'{}'".format(p) for p in inputs)
+            + "\n - ".join(f"'{p}'" for p in inputs)
             + "\nHint: If your path contains spaces, you need to enclose it in quotes")
     input_filename = inputs[0]
     if paired and not interleaved:
@@ -658,7 +667,7 @@ def pipeline_from_parsed_args(args, paired, file_opener, adapters, adapters2) ->
     if paired:
         pair_filter_mode = 'any' if args.pair_filter is None else args.pair_filter
         pipeline = PairedEndPipeline(
-            pair_filter_mode, file_opener
+            file_opener, pair_filter_mode
         )  # type: Any
     else:
         pipeline = SingleEndPipeline(file_opener)
@@ -669,15 +678,14 @@ def pipeline_from_parsed_args(args, paired, file_opener, adapters, adapters2) ->
             args.discard_untrimmed or args.untrimmed_output or args.untrimmed_paired_output):
         pipeline.override_untrimmed_pair_filter = True
 
-    add_unconditional_cutters(pipeline, args.cut, args.cut2, paired)
+    add_unconditional_cutters(pipeline, args.cut, args.cut2)
 
     pipeline_add = pipeline.add_both if paired else pipeline.add
 
     if args.nextseq_trim is not None:
         pipeline_add(NextseqQualityTrimmer(args.nextseq_trim, args.quality_base))
-    if args.quality_cutoff is not None:
-        cutoffs = parse_cutoffs(args.quality_cutoff)
-        pipeline_add(QualityTrimmer(cutoffs[0], cutoffs[1], args.quality_base))
+
+    add_quality_trimmers(pipeline, args.quality_cutoff, args.quality_cutoff2, args.quality_base)
 
     add_adapter_cutter(
         pipeline,
@@ -688,17 +696,18 @@ def pipeline_from_parsed_args(args, paired, file_opener, adapters, adapters2) ->
         args.action,
         args.times,
         args.reverse_complement,
+        not args.rename,  # no "rc" suffix if --rename is used
         args.index,
     )
 
     for modifier in modifiers_applying_to_both_ends_if_paired(args):
         pipeline_add(modifier)
 
-    if args.rename:
-        if args.prefix or args.suffix:
-            raise CommandLineError(
-                "Option --rename cannot be combined with --prefix (-x) or --suffix (-y)"
-            )
+    if args.rename and (args.prefix or args.suffix):
+        raise CommandLineError(
+            "Option --rename cannot be combined with --prefix (-x) or --suffix (-y)"
+        )
+    if args.rename and args.rename != "{header}":
         try:
             if paired:
                 pipeline.add_paired_modifier(PairedEndRenamer(args.rename))
@@ -747,7 +756,7 @@ def adapters_from_args(args) -> Tuple[List[Adapter], List[Adapter]]:
     return adapters, adapters2
 
 
-def add_unconditional_cutters(pipeline: Pipeline, cut1: List[int], cut2: List[int], paired: bool):
+def add_unconditional_cutters(pipeline: Pipeline, cut1: List[int], cut2: List[int]):
     for i, cut_arg in enumerate([cut1, cut2]):
         # cut_arg is a list
         if not cut_arg:
@@ -760,8 +769,7 @@ def add_unconditional_cutters(pipeline: Pipeline, cut1: List[int], cut2: List[in
             if c == 0:
                 continue
             if i == 0:  # R1
-                if paired:
-                    assert isinstance(pipeline, PairedEndPipeline)
+                if isinstance(pipeline, PairedEndPipeline):
                     pipeline.add(UnconditionalCutter(c), None)
                 else:
                     assert isinstance(pipeline, SingleEndPipeline)
@@ -772,6 +780,27 @@ def add_unconditional_cutters(pipeline: Pipeline, cut1: List[int], cut2: List[in
                 pipeline.add(None, UnconditionalCutter(c))
 
 
+def add_quality_trimmers(
+    pipeline: Pipeline,
+    cutoff1: Optional[str],
+    cutoff2: Optional[str],
+    quality_base: int,
+):
+    qtrimmers = [
+        QualityTrimmer(*parse_cutoffs(cutoff), quality_base)
+        if cutoff is not None and cutoff != "0" else None
+        for cutoff in (cutoff1, cutoff2)
+    ]
+    if isinstance(pipeline, PairedEndPipeline):
+        if cutoff1 is not None and cutoff2 is None:
+            qtrimmers[1] = copy.copy(qtrimmers[0])
+        if qtrimmers[0] is not None or qtrimmers[1] is not None:
+            pipeline.add(*qtrimmers)
+    elif qtrimmers[0] is not None:
+        assert isinstance(pipeline, SingleEndPipeline)
+        pipeline.add(qtrimmers[0])
+
+
 def add_adapter_cutter(
     pipeline,
     adapters,
@@ -781,6 +810,7 @@ def add_adapter_cutter(
     action: Optional[str],
     times: int,
     reverse_complement: bool,
+    add_rc_suffix: bool,
     allow_index: bool,
 ):
     if pair_adapters:
@@ -808,7 +838,8 @@ def add_adapter_cutter(
         elif adapter_cutter:
             if reverse_complement:
                 modifier = ReverseComplementer(
-                    adapter_cutter
+                    adapter_cutter,
+                    rc_suffix=" rc" if add_rc_suffix else None,
                 )  # type: Union[AdapterCutter,ReverseComplementer]
             else:
                 modifier = adapter_cutter
@@ -842,6 +873,7 @@ def log_header(cmdlineargs):
 
 def main_cli():  # pragma: no cover
     """Entry point for command-line script"""
+    multiprocessing.freeze_support()
     main(sys.argv[1:])
     return 0
 
@@ -860,13 +892,12 @@ def main(cmdlineargs, default_outfile=sys.stdout.buffer) -> Statistics:
     # Setup logging only if there are not already any handlers (can happen when
     # this function is being called externally such as from unit tests)
     if not logging.root.handlers:
-        # If results are to be sent to stdout, logging needs to go to stderr
-        log_to_stdout = args.output is not None and args.output != "-" and args.paired_output != "-"
-        setup_logging(logger, stdout=log_to_stdout,
+        setup_logging(logger, log_to_stderr=is_any_output_stdout(args),
             quiet=args.quiet, minimal=args.report == 'minimal', debug=args.debug)
     log_header(cmdlineargs)
     profiler = setup_profiler_if_requested(args.profile)
 
+    log_system_info()
     if args.quiet and args.report:
         parser.error("Options --quiet and --report cannot be used at the same time")
 
@@ -879,13 +910,12 @@ def main(cmdlineargs, default_outfile=sys.stdout.buffer) -> Statistics:
 
     cores = available_cpu_count() if args.cores == 0 else args.cores
     file_opener = FileOpener(
-        compression_level=args.compression_level, threads=0 if cores == 1 else None)
-    if sys.stderr.isatty() and not args.quiet:
+        compression_level=args.compression_level, threads=estimate_compression_threads(cores))
+    if sys.stderr.isatty() and not args.quiet and not args.debug:
         progress = Progress()
     else:
         progress = DummyProgress()
     paired = determine_paired(args)
-    assert paired in (False, True)
 
     try:
         is_interleaved_input = args.interleaved and len(args.inputs) == 1
@@ -917,14 +947,26 @@ def main(cmdlineargs, default_outfile=sys.stdout.buffer) -> Statistics:
         sys.exit(1)
     except (dnaio.FileFormatError, dnaio.UnknownFileFormat, EOFError) as e:
         logger.debug("Command line error. Traceback:", exc_info=True)
-        sys.exit("cutadapt: error: {}".format(e))
+        sys.exit(f"cutadapt: error: {e}")
 
     elapsed = time.time() - start_time
     if args.report == 'minimal':
         report = minimal_report
     else:
         report = full_report
-    logger.log(REPORT, '%s', report(stats, elapsed, args.gc_content / 100))
+    logger.log(REPORT, '%s', report(stats, elapsed, args.gc_content / 100.0))
+    if args.json is not None:
+        with open(args.json, "w") as f:
+            json_dict = json_report(
+                stats=stats,
+                cmdlineargs=cmdlineargs,
+                inpaths=inpaths,
+                cores=cores,
+                paired=paired,
+                gc_content=args.gc_content / 100.0,
+            )
+            f.write(json_dumps(json_dict))
+            f.write("\n")
     if profiler is not None:
         import pstats
         profiler.disable()
@@ -932,6 +974,12 @@ def main(cmdlineargs, default_outfile=sys.stdout.buffer) -> Statistics:
     return stats
 
 
+def log_system_info():
+    logger.debug("Python executable: %s", sys.executable)
+    logger.debug("dnaio version: %s", dnaio.__version__)
+    logger.debug("xopen version: %s", xopen.__version__)
+
+
 def setup_runner(
     pipeline: Pipeline,
     inpaths: InputPaths,
@@ -944,7 +992,7 @@ def setup_runner(
     try:
         if cores > 1:
             return ParallelPipelineRunner(
-                pipeline, inpaths, outfiles, progress, n_workers=cores, buffer_size=buffer_size)
+                pipeline, inpaths, outfiles, file_opener, progress, n_workers=cores, buffer_size=buffer_size)
         else:
             infiles = inpaths.open(file_opener)
             return SerialPipelineRunner(pipeline, infiles, outfiles, progress)
@@ -972,5 +1020,51 @@ def warn_if_en_dashes(args):
             )
 
 
+def estimate_compression_threads(cores: int) -> Optional[int]:
+    return max(0, min(cores, 4))
+
+
+def is_any_output_stdout(args):
+    return any([
+        args.output is None,
+        args.output == "-",
+        args.paired_output == "-",
+        args.untrimmed_output == "-",
+        args.untrimmed_paired_output == "-",
+        args.too_short_output == "-",
+        args.too_short_paired_output == "-",
+        args.too_long_output == "-",
+        args.too_long_paired_output == "-",
+        args.rest_file == "-",
+        args.info_file == "-",
+        args.wildcard_file == "-",
+    ])
+
+
+def json_report(
+    stats: Statistics,
+    cmdlineargs: List[str],
+    inpaths: InputPaths,
+    cores: int,
+    paired: bool,
+    gc_content: float,
+) -> Dict:
+    d = {
+        "tag": "Cutadapt report",
+        "schema_version": OneLine([0, 1]),
+        "cutadapt_version": __version__,
+        "python_version": platform.python_version(),
+        "command_line_arguments": cmdlineargs,
+        "cores": cores,
+        "input": {
+            "path1": inpaths.path1,
+            "path2": inpaths.path2,
+            "paired": paired,
+        },
+    }
+    d.update(stats.as_json(gc_content))
+    return d
+
+
 if __name__ == '__main__':  # pragma: no cover
     sys.exit(main_cli())
