@@ -7,7 +7,7 @@ The ...Match classes trim the reads.
 import logging
 from enum import Enum
 from collections import defaultdict
-from typing import Optional, Tuple, Sequence, Dict, Any, List
+from typing import Optional, Tuple, Sequence, Dict, Any, List, Union
 from abc import ABC, abstractmethod
 
 from . import align
@@ -45,15 +45,18 @@ class EndStatistics:
     """Statistics about the 5' or 3' end"""
 
     def __init__(self, adapter: "SingleAdapter"):
-        self.max_error_rate = adapter.max_error_rate  # type: float
-        self.sequence = adapter.sequence  # type: str
-        self.effective_length = adapter.effective_length  # type: int
-        self.has_wildcards = adapter.adapter_wildcards  # type: bool
-        # self.errors[l][e] == n iff n times a sequence of length l matching at e errors was removed
-        self.errors = defaultdict(returns_defaultdict_int)  # type: Dict[int, Dict[int, int]]
+        self.max_error_rate: float = adapter.max_error_rate
+        self.sequence: str = adapter.sequence
+        self.effective_length: int = adapter.effective_length
+        self.has_wildcards: bool = adapter.adapter_wildcards
+        self.indels: bool = adapter.indels
+        self.adapter_type: str = adapter.descriptive_identifier()
+        self.allows_partial_matches: bool = adapter.allows_partial_matches
+        # self.errors[l][e] == n iff a sequence of length l matching at e errors was removed n times
+        self.errors: Dict[int, Dict[int, int]] = defaultdict(returns_defaultdict_int)
         self.adjacent_bases = {'A': 0, 'C': 0, 'G': 0, 'T': 0, '': 0}
         # TODO avoid hard-coding the list of classes
-        self._remove_prefix = isinstance(adapter, (FrontAdapter, NonInternalFrontAdapter, PrefixAdapter))
+        self._remove_prefix = isinstance(adapter, FrontAdapter)
 
     def __repr__(self):
         errors = {k: dict(v) for k, v in self.errors.items()}
@@ -70,6 +73,7 @@ class EndStatistics:
             self.max_error_rate != other.max_error_rate
             or self.sequence != other.sequence
             or self.effective_length != other.effective_length
+            or self.indels != other.indels
         ):
             raise RuntimeError('Incompatible EndStatistics, cannot be added')
         for base in ('A', 'C', 'G', 'T', ''):
@@ -93,6 +97,7 @@ class EndStatistics:
         i bases of this adapter match a random sequence with
         GC content gc_content.
         """
+        assert 0.0 <= gc_content <= 1.0
         seq = self.sequence
         # FIXME this is broken for 'anywhere' adapters
         if self._remove_prefix:
@@ -109,41 +114,152 @@ class EndStatistics:
         return probabilities
 
 
-class AdapterStatistics:
+class AdapterStatistics(ABC):
+
+    reverse_complemented: int = 0
+    name: str
+    adapter: "Adapter"
+
+    @abstractmethod
+    def __iadd__(self, other):
+        pass
+
+    @abstractmethod
+    def end_statistics(self) -> Tuple[Optional[EndStatistics], Optional[EndStatistics]]:
+        pass
+
+    @abstractmethod
+    def add_match(self, match) -> None:
+        pass
+
+
+class SingleAdapterStatistics(AdapterStatistics, ABC):
     """
-    Statistics about an adapter. An adapter can work on the 5' end (front)
-    or 3' end (back) of a read, and statistics for that are captured
-    separately in EndStatistics objects.
+    Statistics about a 5' or 3' adapter, where we only need to keep track of sequences
+    removed from one "end".
+    """
+
+    def __init__(self, adapter: "SingleAdapter"):
+        self.name = adapter.name
+        self.adapter = adapter
+        self.end = EndStatistics(adapter)
+
+    def __repr__(self):
+        return f"SingleAdapterStatistics(name={self.name}, end={self.end})"
+
+    def __iadd__(self, other: "SingleAdapterStatistics"):
+        if not isinstance(other, self.__class__):
+            raise ValueError("Cannot iadd")
+        self.end += other.end
+        self.reverse_complemented += other.reverse_complemented
+        return self
+
+
+class FrontAdapterStatistics(SingleAdapterStatistics):
+
+    def add_match(self, match: "RemoveBeforeMatch"):
+        self.end.errors[match.removed_sequence_length()][match.errors] += 1
+
+    def end_statistics(self) -> Tuple[Optional[EndStatistics], Optional[EndStatistics]]:
+        return self.end, None
+
+
+class BackAdapterStatistics(SingleAdapterStatistics):
+
+    def add_match(self, match: "RemoveAfterMatch"):
+        adjacent_base = match.adjacent_base()
+        self.end.errors[match.removed_sequence_length()][match.errors] += 1
+        try:
+            self.end.adjacent_bases[adjacent_base] += 1
+        except KeyError:
+            self.end.adjacent_bases[""] += 1
+
+    def end_statistics(self) -> Tuple[Optional[EndStatistics], Optional[EndStatistics]]:
+        return None, self.end
+
+
+class LinkedAdapterStatistics(AdapterStatistics):
+    """
+    Statistics about sequences removed by a lined adapter.
     """
 
     def __init__(
         self,
-        adapter: "Adapter",
+        adapter: "LinkedAdapter",
         front: "SingleAdapter",
-        back: Optional["SingleAdapter"] = None,
+        back: "SingleAdapter",
     ):
         self.name = adapter.name
         self.adapter = adapter
         self.front = EndStatistics(front)
-        if back is None:
-            self.back = EndStatistics(front)
-        else:
-            self.back = EndStatistics(back)
+        self.back = EndStatistics(back)
         self.reverse_complemented = 0
 
     def __repr__(self):
-        return "AdapterStatistics(name={}, front={}, back={})".format(
-            self.name,
-            self.front,
-            self.back,
-        )
+        return f"LinkedAdapterStatistics(name={self.name}, front={self.front}, back={self.back})"
 
-    def __iadd__(self, other: "AdapterStatistics"):
+    def __iadd__(self, other: "LinkedAdapterStatistics"):
+        if not isinstance(other, self.__class__):
+            raise ValueError("Cannot iadd")
         self.front += other.front
         self.back += other.back
         self.reverse_complemented += other.reverse_complemented
         return self
 
+    def add_match(self, match: "LinkedMatch"):
+        # TODO this is duplicated code
+        if match.front_match:
+            self.front.errors[match.front_match.removed_sequence_length()][match.errors] += 1
+        if match.back_match:
+            adjacent_base = match.back_match.adjacent_base()
+            self.back.errors[match.back_match.removed_sequence_length()][match.errors] += 1
+            try:
+                self.back.adjacent_bases[adjacent_base] += 1
+            except KeyError:
+                self.back.adjacent_bases[""] += 1
+
+    def end_statistics(self) -> Tuple[Optional[EndStatistics], Optional[EndStatistics]]:
+        return self.front, self.back
+
+
+class AnywhereAdapterStatistics(AdapterStatistics):
+    """
+    Statistics about sequences removed by a lined adapter.
+    """
+
+    def __init__(self, adapter: "AnywhereAdapter"):
+        self.name = adapter.name
+        self.adapter = adapter
+        self.front = EndStatistics(adapter)
+        self.back = EndStatistics(adapter)
+        self.reverse_complemented = 0
+
+    def __repr__(self):
+        return f"AnywhereAdapterStatistics(name={self.name}, front={self.front}, back={self.back})"
+
+    def __iadd__(self, other: "AnywhereAdapterStatistics"):
+        if not isinstance(other, AnywhereAdapterStatistics):
+            raise ValueError("Cannot add")
+        self.front += other.front
+        self.back += other.back
+        self.reverse_complemented += other.reverse_complemented
+        return self
+
+    def add_match(self, match: Union["RemoveBeforeMatch", "RemoveAfterMatch"]) -> None:
+        # TODO contains duplicated code from the other add_match() methods
+        if isinstance(match, RemoveBeforeMatch):
+            self.front.errors[match.removed_sequence_length()][match.errors] += 1
+        else:
+            adjacent_base = match.adjacent_base()
+            self.back.errors[match.removed_sequence_length()][match.errors] += 1
+            try:
+                self.back.adjacent_bases[adjacent_base] += 1
+            except KeyError:
+                self.back.adjacent_bases[""] += 1
+
+    def end_statistics(self) -> Tuple[Optional[EndStatistics], Optional[EndStatistics]]:
+        return self.front, self.back
+
 
 class Match(ABC):
 
@@ -184,19 +300,18 @@ class SingleMatch(Match, ABC):
         adapter: "SingleAdapter",
         sequence: str,
     ):
-        self.adjacent_base = ""
-        self.astart = astart  # type: int
-        self.astop = astop  # type: int
-        self.rstart = rstart  # type: int
-        self.rstop = rstop  # type: int
-        self.matches = matches  # type: int
-        self.errors = errors  # type: int
-        self.adapter = adapter  # type: SingleAdapter
+        self.astart: int = astart
+        self.astop: int = astop
+        self.rstart: int = rstart
+        self.rstop: int = rstop
+        self.matches: int = matches
+        self.errors: int = errors
+        self.adapter: SingleAdapter = adapter
         self.sequence = sequence
         # Number of aligned characters in the adapter. If there are
         # indels, this may be different from the number of characters
         # in the read.
-        self.length = astop - astart  # type: int
+        self.length: int = astop - astart
 
     def __repr__(self):
         return 'SingleMatch(astart={}, astop={}, rstart={}, rstop={}, matches={}, errors={})'.format(
@@ -253,6 +368,10 @@ class SingleMatch(Match, ABC):
 
         return [info]
 
+    @abstractmethod
+    def removed_sequence_length(self) -> int:
+        pass
+
 
 class RemoveBeforeMatch(SingleMatch):
     """A match that removes sequence before the match"""
@@ -287,9 +406,8 @@ class RemoveBeforeMatch(SingleMatch):
     def trimmed(self, read):
         return read[self.rstop:]
 
-    def update_statistics(self, statistics: AdapterStatistics):
-        """Update AdapterStatistics in place"""
-        statistics.front.errors[self.rstop][self.errors] += 1
+    def removed_sequence_length(self) -> int:
+        return self.rstop
 
 
 class RemoveAfterMatch(SingleMatch):
@@ -325,14 +443,11 @@ class RemoveAfterMatch(SingleMatch):
     def trimmed(self, read):
         return read[:self.rstart]
 
-    def update_statistics(self, statistics: AdapterStatistics):
-        """Update AdapterStatistics in place"""
-        adjacent_base = self.sequence[self.rstart - 1:self.rstart]
-        statistics.back.errors[len(self.sequence) - self.rstart][self.errors] += 1
-        try:
-            statistics.back.adjacent_bases[adjacent_base] += 1
-        except KeyError:
-            statistics.back.adjacent_bases[''] = 1
+    def adjacent_base(self) -> str:
+        return self.sequence[self.rstart - 1:self.rstart]
+
+    def removed_sequence_length(self) -> int:
+        return len(self.sequence) - self.rstart
 
 
 def _generate_adapter_name(_start=[1]) -> str:
@@ -360,10 +475,18 @@ class Adapter(Matchable, ABC):
 
     description = "adapter with one component"  # this is overriden in subclasses
 
+    @abstractmethod
+    def spec(self) -> str:
+        """Return string representation of this adapter"""
+
     @abstractmethod
     def create_statistics(self) -> AdapterStatistics:
         pass
 
+    @abstractmethod
+    def descriptive_identifier(self) -> str:
+        pass
+
 
 class SingleAdapter(Adapter, ABC):
     """
@@ -395,6 +518,8 @@ class SingleAdapter(Adapter, ABC):
         unique number.
     """
 
+    allows_partial_matches: bool = True
+
     def __init__(
         self,
         sequence: str,
@@ -405,27 +530,33 @@ class SingleAdapter(Adapter, ABC):
         name: Optional[str] = None,
         indels: bool = True,
     ):
-        self.name = _generate_adapter_name() if name is None else name  # type: str
+        self.name: str = _generate_adapter_name() if name is None else name
         super().__init__(self.name)
-        self._debug = False  # type: bool
-        self.sequence = sequence.upper().replace("U", "T")  # type: str
+        self._debug: bool = False
+        self.sequence: str = sequence.upper().replace("U", "T")
         if not self.sequence:
             raise ValueError("Adapter sequence is empty")
         if max_errors >= 1:
             max_errors /= len(self.sequence)
-        self.max_error_rate = max_errors  # type: float
-        self.min_overlap = min(min_overlap, len(self.sequence))  # type: int
+        self.max_error_rate: float = max_errors
+        self.min_overlap: int = min(min_overlap, len(self.sequence))
         iupac = frozenset('ABCDGHKMNRSTUVWXY')
         if adapter_wildcards and not set(self.sequence) <= iupac:
             for c in self.sequence:
                 if c not in iupac:
-                    raise InvalidCharacter('Character {!r} in adapter sequence {!r} is '
-                        'not a valid IUPAC code. Use only characters '
-                        'ABCDGHKMNRSTUVWXY.'.format(c, self.sequence))
+                    if c == "I":
+                        extra = "For inosine, consider using N instead and please comment " \
+                                "on <https://github.com/marcelm/cutadapt/issues/546>."
+                    else:
+                        extra = "Use only characters 'ABCDGHKMNRSTUVWXY'."
+                    raise InvalidCharacter(
+                        f"Character '{c}' in adapter sequence '{self.sequence}' is "
+                        f"not a valid IUPAC code. {extra}"
+                    )
         # Optimization: Use non-wildcard matching if only ACGT is used
-        self.adapter_wildcards = adapter_wildcards and not set(self.sequence) <= set("ACGT")  # type: bool
-        self.read_wildcards = read_wildcards  # type: bool
-        self.indels = indels  # type: bool
+        self.adapter_wildcards: bool = adapter_wildcards and not set(self.sequence) <= set("ACGT")
+        self.read_wildcards: bool = read_wildcards
+        self.indels: bool = indels
         self.aligner = self._aligner()
 
     def _make_aligner(self, flags: int) -> align.Aligner:
@@ -479,9 +610,6 @@ class SingleAdapter(Adapter, ABC):
     def __len__(self) -> int:
         return len(self.sequence)
 
-    def create_statistics(self) -> AdapterStatistics:
-        return AdapterStatistics(self, self)
-
 
 class FrontAdapter(SingleAdapter):
     """A 5' adapter"""
@@ -492,6 +620,9 @@ class FrontAdapter(SingleAdapter):
         self._force_anywhere = kwargs.pop("force_anywhere", False)
         super().__init__(*args, **kwargs)
 
+    def descriptive_identifier(self) -> str:
+        return "regular_five_prime"
+
     def _aligner(self) -> align.Aligner:
         return self._make_aligner(Where.ANYWHERE.value if self._force_anywhere else Where.FRONT.value)
 
@@ -503,13 +634,19 @@ class FrontAdapter(SingleAdapter):
         return None if no match was found given the matching criteria (minimum
         overlap length, maximum error rate).
         """
-        alignment = self.aligner.locate(sequence)  # type: Optional[Tuple[int,int,int,int,int,int]]
+        alignment: Optional[Tuple[int, int, int, int, int, int]] = self.aligner.locate(sequence)
         if self._debug:
             print(self.aligner.dpmatrix)
         if alignment is None:
             return None
         return RemoveBeforeMatch(*alignment, adapter=self, sequence=sequence)
 
+    def spec(self) -> str:
+        return f"{self.sequence}..."
+
+    def create_statistics(self) -> FrontAdapterStatistics:
+        return FrontAdapterStatistics(self)
+
 
 class BackAdapter(SingleAdapter):
     """A 3' adapter"""
@@ -520,6 +657,9 @@ class BackAdapter(SingleAdapter):
         self._force_anywhere = kwargs.pop("force_anywhere", False)
         super().__init__(*args, **kwargs)
 
+    def descriptive_identifier(self) -> str:
+        return "regular_three_prime"
+
     def _aligner(self):
         return self._make_aligner(Where.ANYWHERE.value if self._force_anywhere else Where.BACK.value)
 
@@ -531,13 +671,19 @@ class BackAdapter(SingleAdapter):
         return None if no match was found given the matching criteria (minimum
         overlap length, maximum error rate).
         """
-        alignment = self.aligner.locate(sequence)  # type: Optional[Tuple[int,int,int,int,int,int]]
+        alignment: Optional[Tuple[int, int, int, int, int, int]] = self.aligner.locate(sequence)
         if self._debug:
             print(self.aligner.dpmatrix)  # pragma: no cover
         if alignment is None:
             return None
         return RemoveAfterMatch(*alignment, adapter=self, sequence=sequence)
 
+    def spec(self) -> str:
+        return f"{self.sequence}"
+
+    def create_statistics(self) -> BackAdapterStatistics:
+        return BackAdapterStatistics(self)
+
 
 class AnywhereAdapter(SingleAdapter):
     """
@@ -547,6 +693,9 @@ class AnywhereAdapter(SingleAdapter):
 
     description = "variable 5'/3'"
 
+    def descriptive_identifier(self) -> str:
+        return "anywhere"
+
     def _aligner(self):
         return self._make_aligner(Where.ANYWHERE.value)
 
@@ -570,12 +719,21 @@ class AnywhereAdapter(SingleAdapter):
             match = RemoveAfterMatch(*alignment, adapter=self, sequence=sequence)  # type: ignore
         return match
 
+    def spec(self) -> str:
+        return f"...{self.sequence}..."
+
+    def create_statistics(self) -> AnywhereAdapterStatistics:
+        return AnywhereAdapterStatistics(self)
+
 
 class NonInternalFrontAdapter(FrontAdapter):
     """A non-internal 5' adapter"""
 
     description = "non-internal 5'"
 
+    def descriptive_identifier(self) -> str:
+        return "noninternal_five_prime"
+
     def _aligner(self):
         return self._make_aligner(Where.FRONT_NOT_INTERNAL.value)
 
@@ -591,12 +749,18 @@ class NonInternalFrontAdapter(FrontAdapter):
             return None
         return RemoveBeforeMatch(*alignment, adapter=self, sequence=sequence)  # type: ignore
 
+    def spec(self) -> str:
+        return f"X{self.sequence}..."
+
 
 class NonInternalBackAdapter(BackAdapter):
     """A non-internal 3' adapter"""
 
     description = "non-internal 3'"
 
+    def descriptive_identifier(self) -> str:
+        return "noninternal_three_prime"
+
     def _aligner(self):
         return self._make_aligner(Where.BACK_NOT_INTERNAL.value)
 
@@ -612,11 +776,22 @@ class NonInternalBackAdapter(BackAdapter):
             return None
         return RemoveAfterMatch(*alignment, adapter=self, sequence=sequence)  # type: ignore
 
+    def spec(self) -> str:
+        return f"{self.sequence}X"
+
 
 class PrefixAdapter(NonInternalFrontAdapter):
     """An anchored 5' adapter"""
 
     description = "anchored 5'"
+    allows_partial_matches = False
+
+    def __init__(self, sequence: str, *args, **kwargs):
+        kwargs["min_overlap"] = len(sequence)
+        super().__init__(sequence, *args, **kwargs)
+
+    def descriptive_identifier(self) -> str:
+        return "anchored_five_prime"
 
     def _aligner(self):
         if not self.indels:  # TODO or if error rate allows 0 errors anyway
@@ -630,11 +805,22 @@ class PrefixAdapter(NonInternalFrontAdapter):
         else:
             return self._make_aligner(Where.PREFIX.value)
 
+    def spec(self) -> str:
+        return f"^{self.sequence}..."
+
 
 class SuffixAdapter(NonInternalBackAdapter):
     """An anchored 3' adapter"""
 
     description = "anchored 3'"
+    allows_partial_matches = False
+
+    def __init__(self, sequence: str, *args, **kwargs):
+        kwargs["min_overlap"] = len(sequence)
+        super().__init__(sequence, *args, **kwargs)
+
+    def descriptive_identifier(self) -> str:
+        return "anchored_three_prime"
 
     def _aligner(self):
         if not self.indels:  # TODO or if error rate allows 0 errors anyway
@@ -648,6 +834,9 @@ class SuffixAdapter(NonInternalBackAdapter):
         else:
             return self._make_aligner(Where.SUFFIX.value)
 
+    def spec(self) -> str:
+        return f"{self.sequence}$"
+
 
 class LinkedMatch(Match):
     """
@@ -655,9 +844,9 @@ class LinkedMatch(Match):
     """
     def __init__(self, front_match: RemoveBeforeMatch, back_match: RemoveAfterMatch, adapter: "LinkedAdapter"):
         assert front_match is not None or back_match is not None
-        self.front_match = front_match  # type: RemoveBeforeMatch
-        self.back_match = back_match  # type: RemoveAfterMatch
-        self.adapter = adapter  # type: LinkedAdapter
+        self.front_match: RemoveBeforeMatch = front_match
+        self.back_match: RemoveAfterMatch = back_match
+        self.adapter: LinkedAdapter = adapter
 
     def __repr__(self):
         return '<LinkedMatch(front_match={!r}, back_match={}, adapter={})>'.format(
@@ -689,18 +878,6 @@ class LinkedMatch(Match):
             read = self.back_match.trimmed(read)
         return read
 
-    @property
-    def adjacent_base(self):
-        return self.back_match.adjacent_base
-
-    def update_statistics(self, statistics):
-        """Update AdapterStatistics in place"""
-        if self.front_match:
-            statistics.front.errors[self.front_match.rstop][self.front_match.errors] += 1
-        if self.back_match:
-            length = len(self.back_match.sequence) - self.back_match.rstart
-            statistics.back.errors[length][self.back_match.errors] += 1
-
     def remainder_interval(self) -> Tuple[int, int]:
         matches = [match for match in [self.front_match, self.back_match] if match is not None]
         return remainder(matches)
@@ -756,6 +933,9 @@ class LinkedAdapter(Adapter):
         self.front_adapter.name = self.name
         self.back_adapter = back_adapter
 
+    def descriptive_identifier(self) -> str:
+        return "linked"
+
     def enable_debug(self):
         self.front_adapter.enable_debug()
         self.back_adapter.enable_debug()
@@ -774,8 +954,8 @@ class LinkedAdapter(Adapter):
             return None
         return LinkedMatch(front_match, back_match, self)
 
-    def create_statistics(self) -> AdapterStatistics:
-        return AdapterStatistics(self, self.front_adapter, self.back_adapter)
+    def create_statistics(self) -> LinkedAdapterStatistics:
+        return LinkedAdapterStatistics(self, front=self.front_adapter, back=self.back_adapter)
 
     @property
     def sequence(self):
@@ -785,6 +965,9 @@ class LinkedAdapter(Adapter):
     def remove(self):
         return None
 
+    def spec(self) -> str:
+        return f"{self.front_adapter.spec()}...{self.back_adapter.spec()}"
+
 
 class MultipleAdapters(Matchable):
     """
@@ -858,7 +1041,7 @@ class IndexedAdapters(Matchable, ABC):
         self._make_affix = self._get_make_affix()
 
     def __repr__(self):
-        return "{}(adapters={!r})".format(self.__class__.__name__, self._adapters)
+        return f"{self.__class__.__name__}(adapters={self._adapters!r})"
 
     def match_to(self, sequence: str):
         """Never called because it gets overwritten in __init__"""
@@ -898,7 +1081,7 @@ class IndexedAdapters(Matchable, ABC):
 
     def _make_index(self) -> Tuple[List[int], "AdapterIndex"]:
         logger.info('Building index of %s adapters ...', len(self._adapters))
-        index = dict()  # type: IndexedAdapters.AdapterIndex
+        index: Dict[str, Tuple[SingleAdapter, int, int]] = dict()
         lengths = set()
         has_warned = False
         for adapter in self._adapters:
@@ -949,7 +1132,7 @@ class IndexedAdapters(Matchable, ABC):
         affix = sequence.upper()
 
         # Check all the prefixes or suffixes (affixes) that could match
-        best_adapter = None  # type: Optional[SingleAdapter]
+        best_adapter: Optional[SingleAdapter] = None
         best_length = 0
         best_m = -1
         best_e = 1000
