@@ -4,10 +4,10 @@ A modifier must be callable and typically implemented as a class with a
 __call__ method.
 """
 import re
+import logging
 from types import SimpleNamespace
 from typing import Sequence, List, Tuple, Optional, Set
 from abc import ABC, abstractmethod
-from collections import OrderedDict
 
 from dnaio import record_names_match, Sequence as DnaSequence
 
@@ -17,6 +17,8 @@ from .adapters import MultipleAdapters, SingleAdapter, IndexedPrefixAdapters, In
 from .tokenizer import tokenize_braces, TokenizeError, Token, BraceToken
 from .utils import reverse_complemented_sequence
 
+logger = logging.getLogger()
+
 
 # If the number of prefix or suffix adapters is higher than this, switch to using an index
 INDEXING_THRESHOLD = 5
@@ -37,6 +39,16 @@ class ModificationInfo:
         self.cut_suffix = None
         self.is_rc = None
 
+    def __repr__(self):
+        return (
+            "ModificationInfo("
+            f"matches={self.matches!r}, "
+            f"original_read={self.original_read}, "
+            f"cut_prefix={self.cut_prefix}, "
+            f"cut_suffix={self.cut_suffix}, "
+            f"is_rc={self.is_rc})"
+        )
+
 
 class SingleEndModifier(ABC):
     @abstractmethod
@@ -66,8 +78,8 @@ class PairedEndModifierWrapper(PairedEndModifier):
             raise ValueError("Not both modifiers may be None")
 
     def __repr__(self):
-        return "PairedEndModifierWrapper(modifier1={!r}, modifier2={!r})".format(
-            self._modifier1, self._modifier2)
+        return "PairedEndModifierWrapper(" \
+            f"modifier1={self._modifier1!r}, modifier2={self._modifier2!r})"
 
     def __call__(self, read1, read2, info1: ModificationInfo, info2: ModificationInfo):
         if self._modifier1 is None:
@@ -86,7 +98,7 @@ class AdapterCutter(SingleEndModifier):
 
     def __init__(
         self,
-        adapters: List[Adapter],
+        adapters: Sequence[Adapter],
         times: int = 1,
         action: Optional[str] = "trim",
         index: bool = True,
@@ -105,7 +117,7 @@ class AdapterCutter(SingleEndModifier):
         assert action in ("trim", "mask", "lowercase", "retain", None)
         self.action = action
         self.with_adapters = 0
-        self.adapter_statistics = OrderedDict((a, a.create_statistics()) for a in adapters)
+        self.adapter_statistics = {a: a.create_statistics() for a in adapters}
         if index:
             self.adapters = MultipleAdapters(self._regroup_into_indexed_adapters(adapters))
         else:
@@ -114,8 +126,8 @@ class AdapterCutter(SingleEndModifier):
             raise ValueError("'retain' cannot be combined with times > 1")
 
     def __repr__(self):
-        return 'AdapterCutter(adapters={!r}, times={}, action={!r})'.format(
-            self.adapters, self.times, self.action)
+        return "AdapterCutter(" \
+            f"adapters={self.adapters!r}, times={self.times}, action='{self.action}')"
 
     def _regroup_into_indexed_adapters(self, adapters):
         prefix, suffix, single = self._split_adapters(adapters)
@@ -189,7 +201,7 @@ class AdapterCutter(SingleEndModifier):
         if matches:
             self.with_adapters += 1
             for match in matches:
-                match.update_statistics(self.adapter_statistics[match.adapter])
+                self.adapter_statistics[match.adapter].add_match(match)
         info.matches.extend(matches)  # TODO extend or overwrite?
         return trimmed_read
 
@@ -247,6 +259,9 @@ class ReverseComplementer(SingleEndModifier):
         self.reverse_complemented = 0
         self._suffix = rc_suffix
 
+    def __repr__(self):
+        return f"ReverseComplementer(adapter_cutter={self.adapter_cutter})"
+
     def __call__(self, read, info: ModificationInfo):
         reverse_read = reverse_complemented_sequence(read)
 
@@ -272,7 +287,7 @@ class ReverseComplementer(SingleEndModifier):
             self.adapter_cutter.with_adapters += 1
             for match in matches:
                 stats = self.adapter_cutter.adapter_statistics[match.adapter]
-                match.update_statistics(stats)
+                stats.add_match(match)
                 stats.reverse_complemented += bool(use_reverse_complement)
             info.matches.extend(matches)  # TODO extend or overwrite?
         return trimmed_read
@@ -284,13 +299,13 @@ class PairedAdapterCutterError(Exception):
 
 class PairedAdapterCutter(PairedEndModifier):
     """
-    A Modifier that trims adapter pairs from R1 and R2.
+    Trim adapters in pairs from R1 and R2.
     """
 
     def __init__(self, adapters1, adapters2, action='trim'):
         """
         adapters1 -- list of Adapters to be removed from R1
-        adapters2 -- list of Adapters to be removed from R1
+        adapters2 -- list of Adapters to be removed from R2
 
         Both lists must have the same, non-zero length.
          read pair is trimmed if adapters1[i] is found in R1 and adapters2[i] in R2.
@@ -300,35 +315,30 @@ class PairedAdapterCutter(PairedEndModifier):
         super().__init__()
         if len(adapters1) != len(adapters2):
             raise PairedAdapterCutterError(
-                "The number of reads to trim from R1 and R2 must be the same. "
+                "The number of adapters to trim from R1 and R2 must be the same. "
                 "Given: {} for R1, {} for R2".format(len(adapters1), len(adapters2)))
         if not adapters1:
             raise PairedAdapterCutterError("No adapters given")
-        self._adapters1 = MultipleAdapters(adapters1)
-        self._adapter_indices = {a: i for i, a in enumerate(adapters1)}
-        self._adapters2 = MultipleAdapters(adapters2)
+        self._adapter_pairs = list(zip(adapters1, adapters2))
+        logger.debug("Adapter pairs:")
+        for a1, a2 in self._adapter_pairs:
+            logger.debug(" • %s=%s -- %s=%s", a1.name, a1.spec(), a2.name, a2.spec())
         self.action = action
         self.with_adapters = 0
         self.adapter_statistics = [None, None]
-        self.adapter_statistics[0] = OrderedDict((a, a.create_statistics()) for a in adapters1)
-        self.adapter_statistics[1] = OrderedDict((a, a.create_statistics()) for a in adapters2)
+        self.adapter_statistics[0] = {a: a.create_statistics() for a in adapters1}
+        self.adapter_statistics[1] = {a: a.create_statistics() for a in adapters2}
 
     def __repr__(self):
-        return 'PairedAdapterCutter(adapters1={!r}, adapters2={!r})'.format(
-            self._adapters1, self._adapters2)
+        return f"PairedAdapterCutter(adapter_pairs={self._adapter_pairs!r})"
 
     def __call__(self, read1, read2, info1, info2):
         """
         """
-        match1 = self._adapters1.match_to(read1.sequence)
-        if match1 is None:
+        best_matches = self._find_best_match_pair(read1.sequence, read2.sequence)
+        if best_matches is None:
             return read1, read2
-        adapter1 = match1.adapter
-        adapter2 = self._adapters2[self._adapter_indices[adapter1]]
-        match2 = adapter2.match_to(read2.sequence)
-        if match2 is None:
-            return read1, read2
-
+        match1, match2 = best_matches
         self.with_adapters += 1
         result = []
         for i, match, read in zip([0, 1], [match1, match2], [read1, read2]):
@@ -337,7 +347,7 @@ class PairedAdapterCutter(PairedEndModifier):
                 trimmed_read.sequence = trimmed_read.sequence.upper()
 
             trimmed_read = match.trimmed(trimmed_read)
-            match.update_statistics(self.adapter_statistics[i][match.adapter])
+            self.adapter_statistics[i][match.adapter].add_match(match)
 
             if self.action == 'trim':
                 # read is already trimmed, nothing to do
@@ -356,6 +366,29 @@ class PairedAdapterCutter(PairedEndModifier):
         info2.matches.append(match2)
         return result
 
+    def _find_best_match_pair(self, sequence1: str, sequence2: str) -> Optional[Tuple[Match, Match]]:
+        best = None
+        best_matches = None
+        best_errors = None
+        for adapter1, adapter2 in self._adapter_pairs:
+            match1 = adapter1.match_to(sequence1)
+            if match1 is None:
+                continue
+            match2 = adapter2.match_to(sequence2)
+            if match2 is None:
+                continue
+            total_matches = match1.matches + match2.matches
+            total_errors = match1.errors + match2.errors
+            if (
+                best is None
+                or total_matches > best_matches
+                or (total_matches == best_matches and total_errors < best_errors)
+            ):
+                best = match1, match2
+                best_matches = total_matches
+                best_errors = total_errors
+        return best
+
 
 class UnconditionalCutter(SingleEndModifier):
     """
@@ -367,6 +400,9 @@ class UnconditionalCutter(SingleEndModifier):
     def __init__(self, length: int):
         self.length = length
 
+    def __repr__(self):
+        return f"UnconditionalCutter(length={self.length})"
+
     def __call__(self, read, info: ModificationInfo):
         if self.length > 0:
             info.cut_prefix = read.sequence[:self.length]
@@ -384,6 +420,9 @@ class LengthTagModifier(SingleEndModifier):
         self.regex = re.compile(r"\b" + length_tag + r"[0-9]*\b")
         self.length_tag = length_tag
 
+    def __repr__(self):
+        return f"LengthTagModifier(length_tag='{self.length_tag}')"
+
     def __call__(self, read, info: ModificationInfo):
         read = read[:]
         if read.name.find(self.length_tag) >= 0:
@@ -398,6 +437,9 @@ class SuffixRemover(SingleEndModifier):
     def __init__(self, suffix):
         self.suffix = suffix
 
+    def __repr__(self):
+        return f"SuffixRemover('{self.suffix}')"
+
     def __call__(self, read, info: ModificationInfo):
         read = read[:]
         if read.name.endswith(self.suffix):
@@ -413,6 +455,9 @@ class PrefixSuffixAdder(SingleEndModifier):
         self.prefix = prefix
         self.suffix = suffix
 
+    def __repr__(self):
+        return f"PrefixSuffixAdder(prefix='{self.prefix}', suffix='{self.suffix}')"
+
     def __call__(self, read, info):
         read = read[:]
         adapter_name = info.matches[-1].adapter.name if info.matches else 'no_adapter'
@@ -437,6 +482,7 @@ class Renamer(SingleEndModifier):
     - {cut_prefix} -- prefix removed by UnconditionalCutter (with positive length argument)
     - {cut_suffix} -- suffix removed by UnconditionalCutter (with negative length argument)
     - {adapter_name} -- name of the *last* adapter match or no_adapter if there was none
+    - {rc} -- the string 'rc' if the read was reverse complemented (with --revcomp) or '' otherwise
     """
     variables = {
         "header",
@@ -445,16 +491,20 @@ class Renamer(SingleEndModifier):
         "cut_prefix",
         "cut_suffix",
         "adapter_name",
+        "rc",
     }
 
     def __init__(self, template: str):
         try:
             self._tokens = list(tokenize_braces(template))
         except TokenizeError as e:
-            raise InvalidTemplate("Error in template '{}': {}".format(template, e))
+            raise InvalidTemplate(f"Error in template '{template}': {e}")
         self.raise_if_invalid_variable(self._tokens, self.variables)
         self._template = template
 
+    def __repr__(self):
+        return f"{self.__class__.__name__}('{self._template}')"
+
     @staticmethod
     def raise_if_invalid_variable(tokens: List[Token], allowed: Set[str]) -> None:
         for token in tokens:
@@ -463,7 +513,7 @@ class Renamer(SingleEndModifier):
             value = token.value
             if value not in allowed:
                 raise InvalidTemplate(
-                    "Error in template: Variable '{}' not recognized".format(value)
+                    f"Error in template: Variable '{value}' not recognized"
                 )
 
     @staticmethod
@@ -484,6 +534,7 @@ class Renamer(SingleEndModifier):
             cut_prefix=info.cut_prefix if info.cut_prefix else "",
             cut_suffix=info.cut_suffix if info.cut_suffix else "",
             adapter_name=info.matches[-1].adapter.name if info.matches else "no_adapter",
+            rc="rc" if info.is_rc else "",
         )
         return read
 
@@ -506,15 +557,14 @@ class PairedEndRenamer(PairedEndModifier):
         try:
             self._tokens = list(tokenize_braces(template))
         except TokenizeError as e:
-            raise InvalidTemplate("Error in template '{}': {}".format(template, e))
+            raise InvalidTemplate(f"Error in template '{template}': {e}")
         Renamer.raise_if_invalid_variable(self._tokens, self._get_allowed_variables())
         self._template = template
 
     @staticmethod
     def _get_allowed_variables() -> Set[str]:
-        allowed = Renamer.variables.copy()
-        allowed.add("rn")
-        for v in Renamer.variables - {"id"}:
+        allowed = (Renamer.variables - {"rc"}) | {"rn"}
+        for v in Renamer.variables - {"id", "rc"}:
             allowed.add("r1." + v)
             allowed.add("r2." + v)
         return allowed
@@ -526,7 +576,7 @@ class PairedEndRenamer(PairedEndModifier):
         id1, comment1 = Renamer.parse_name(read1.name)
         id2, comment2 = Renamer.parse_name(read2.name)
         if not record_names_match(read1.name, read2.name):
-            raise ValueError("Input read IDs not identical: '{}' != '{}'".format(id1, id2))
+            raise ValueError(f"Input read IDs not identical: '{id1}' != '{id2}'")
         name1, name2 = self.get_new_headers(
             id1=id1,
             id2=id2,
@@ -594,9 +644,13 @@ class ZeroCapper(SingleEndModifier):
     Change negative quality values of a read to zero
     """
     def __init__(self, quality_base=33):
+        self.quality_base = quality_base
         qb = quality_base
         self.zero_cap_trans = str.maketrans(''.join(map(chr, range(qb))), chr(qb) * qb)
 
+    def __repr__(self):
+        return f"ZeroCapper(quality_base={self.quality_base})"
+
     def __call__(self, read, info: ModificationInfo):
         read = read[:]
         read.qualities = read.qualities.translate(self.zero_cap_trans)
@@ -609,6 +663,9 @@ class NextseqQualityTrimmer(SingleEndModifier):
         self.base = base
         self.trimmed_bases = 0
 
+    def __repr__(self):
+        return f"NextseqQualityTrimmer(cutoff={self.cutoff}, base={self.base})"
+
     def __call__(self, read, info: ModificationInfo):
         stop = nextseq_trim_index(read, self.cutoff, self.base)
         self.trimmed_bases += len(read) - stop
@@ -622,6 +679,10 @@ class QualityTrimmer(SingleEndModifier):
         self.base = base
         self.trimmed_bases = 0
 
+    def __repr__(self):
+        return f"QualityTrimmer(cutoff_front={self.cutoff_front}, " \
+               f"cutoff_back={self.cutoff_back}, base={self.base})"
+
     def __call__(self, read, info: ModificationInfo):
         start, stop = quality_trim_index(read.qualities, self.cutoff_front, self.cutoff_back, self.base)
         self.trimmed_bases += len(read) - (stop - start)
@@ -637,6 +698,9 @@ class Shortener(SingleEndModifier):
     def __init__(self, length):
         self.length = length
 
+    def __repr__(self):
+        return f"Shortener(length={self.length})"
+
     def __call__(self, read, info: ModificationInfo):
         if self.length >= 0:
             return read[:self.length]
@@ -650,6 +714,9 @@ class NEndTrimmer(SingleEndModifier):
         self.start_trim = re.compile(r'^N+')
         self.end_trim = re.compile(r'N+$')
 
+    def __repr__(self):
+        return "NEndTrimmer()"
+
     def __call__(self, read, info: ModificationInfo):
         sequence = read.sequence
         start_cut = self.start_trim.match(sequence)
