@@ -3,7 +3,6 @@ import os
 import sys
 import copy
 import logging
-import functools
 from typing import List, Optional, BinaryIO, TextIO, Any, Tuple, Dict
 from abc import ABC, abstractmethod
 from multiprocessing import Process, Pipe, Queue
@@ -12,17 +11,16 @@ import multiprocessing.connection
 from multiprocessing.connection import Connection
 import traceback
 
-from xopen import xopen
 import dnaio
 
 from .utils import Progress, FileOpener
 from .modifiers import SingleEndModifier, PairedEndModifier, PairedEndModifierWrapper, ModificationInfo
 from .report import Statistics
-from .filters import (Redirector, PairedRedirector, NoFilter, PairedNoFilter, InfoFileWriter,
-    RestFileWriter, WildcardFileWriter, TooShortReadFilter, TooLongReadFilter, NContentFilter,
-    MaximumExpectedErrorsFilter,
-    CasavaFilter, DiscardTrimmedFilter, DiscardUntrimmedFilter, Demultiplexer,
-    PairedDemultiplexer, CombinatorialDemultiplexer)
+from .filters import (TooShort, TooLong, TooManyN, TooManyExpectedErrors, CasavaFiltered,
+    DiscardTrimmed, DiscardUntrimmed, Predicate)
+from .steps import SingleEndSink, PairedEndSink, SingleEndFilter, PairedEndFilter, Demultiplexer, \
+    PairedDemultiplexer, CombinatorialDemultiplexer, RestFileWriter, WildcardFileWriter, \
+    InfoFileWriter, SingleEndStep, PairedSingleEndStep
 
 logger = logging.getLogger()
 
@@ -57,7 +55,7 @@ class OutputFiles:
     """
     The attributes are either None or open file-like objects except for demultiplex_out
     and demultiplex_out2, which are dictionaries that map an adapter name
-    to file-like objects.
+    to a file-like object.
     """
     def __init__(
         self,
@@ -155,7 +153,7 @@ class Pipeline(ABC):
 
     def __init__(self, file_opener: FileOpener):
         self._reader = None  # type: Any
-        self._filters = []  # type: List[Any]
+        self._steps = []  # type: List[Any]
         self._infiles = None  # type: Optional[InputFiles]
         self._outfiles = None  # type: Optional[OutputFiles]
         self._demultiplexer = None
@@ -186,11 +184,11 @@ class Pipeline(ABC):
         pass
 
     def _set_output(self, outfiles: OutputFiles) -> None:
-        self._filters = []
+        self._steps = []
+        self._textiowrappers = []
         self._outfiles = outfiles
-        filter_wrapper = self._filter_wrapper()
 
-        for filter_class, outfile in (
+        for step_class, outfile in (
             (RestFileWriter, outfiles.rest),
             (InfoFileWriter, outfiles.info),
             (WildcardFileWriter, outfiles.wildcard),
@@ -198,37 +196,38 @@ class Pipeline(ABC):
             if outfile:
                 textiowrapper = io.TextIOWrapper(outfile)
                 self._textiowrappers.append(textiowrapper)
-                self._filters.append(filter_wrapper(None, filter_class(textiowrapper), None))
+                self._steps.append(self._wrap_single_end_step(step_class(textiowrapper)))
 
         # minimum length and maximum length
-        for lengths, file1, file2, filter_class in (
-                (self._minimum_length, outfiles.too_short, outfiles.too_short2, TooShortReadFilter),
-                (self._maximum_length, outfiles.too_long, outfiles.too_long2, TooLongReadFilter)
+        for lengths, file1, file2, predicate_class in (
+            (self._minimum_length, outfiles.too_short, outfiles.too_short2, TooShort),
+            (self._maximum_length, outfiles.too_long, outfiles.too_long2, TooLong)
         ):
             if lengths is None:
                 continue
             writer = self._open_writer(file1, file2) if file1 else None
-            f1 = filter_class(lengths[0]) if lengths[0] is not None else None
+            f1 = predicate_class(lengths[0]) if lengths[0] is not None else None
             if len(lengths) == 2 and lengths[1] is not None:
-                f2 = filter_class(lengths[1])
+                f2 = predicate_class(lengths[1])
             else:
                 f2 = None
-            self._filters.append(filter_wrapper(writer, filter=f1, filter2=f2))
+            self._steps.append(self._make_filter(writer, predicate1=f1, predicate2=f2))
 
         if self.max_n is not None:
-            f1 = f2 = NContentFilter(self.max_n)
-            self._filters.append(filter_wrapper(None, f1, f2))
+            f1 = f2 = TooManyN(self.max_n)
+            self._steps.append(self._make_filter(None, f1, f2))
 
         if self.max_expected_errors is not None:
             if not self._reader.delivers_qualities:
-                logger.warning("Ignoring option --max-ee as input does not contain quality values")
+                logger.warning(
+                    "Ignoring option --max-ee because input does not contain quality values")
             else:
-                f1 = f2 = MaximumExpectedErrorsFilter(self.max_expected_errors)
-                self._filters.append(filter_wrapper(None, f1, f2))
+                f1 = f2 = TooManyExpectedErrors(self.max_expected_errors)
+                self._steps.append(self._make_filter(None, f1, f2))
 
         if self.discard_casava:
-            f1 = f2 = CasavaFilter()
-            self._filters.append(filter_wrapper(None, f1, f2))
+            f1 = f2 = CasavaFiltered()
+            self._steps.append(self._make_filter(None, f1, f2))
 
         if int(self.discard_trimmed) + int(self.discard_untrimmed) + int(outfiles.untrimmed is not None) > 1:
             raise ValueError('discard_trimmed, discard_untrimmed and outfiles.untrimmed must not '
@@ -236,26 +235,28 @@ class Pipeline(ABC):
 
         if outfiles.demultiplex_out is not None or outfiles.combinatorial_out is not None:
             self._demultiplexer = self._create_demultiplexer(outfiles)
-            self._filters.append(self._demultiplexer)
+            self._steps.append(self._demultiplexer)
         else:
-            # Allow overriding the wrapper for --discard-untrimmed/--untrimmed-(paired-)output
-            untrimmed_filter_wrapper = self._untrimmed_filter_wrapper()
+            # Some special handling to allow overriding the wrapper for
+            # --discard-untrimmed/--untrimmed-(paired-)output
 
             # Set up the remaining filters to deal with --discard-trimmed,
             # --discard-untrimmed and --untrimmed-output. These options
             # are mutually exclusive in order to avoid brain damage.
             if self.discard_trimmed:
-                self._filters.append(
-                    filter_wrapper(None, DiscardTrimmedFilter(), DiscardTrimmedFilter()))
+                self._steps.append(
+                    self._make_filter(None, DiscardTrimmed(), DiscardTrimmed()))
             elif self.discard_untrimmed:
-                self._filters.append(
-                    untrimmed_filter_wrapper(None, DiscardUntrimmedFilter(), DiscardUntrimmedFilter()))
+                self._steps.append(
+                    self._make_untrimmed_filter(None))
             elif outfiles.untrimmed:
                 untrimmed_writer = self._open_writer(outfiles.untrimmed, outfiles.untrimmed2)
-                self._filters.append(
-                    untrimmed_filter_wrapper(untrimmed_writer, DiscardUntrimmedFilter(), DiscardUntrimmedFilter()))
-            self._filters.append(self._final_filter(outfiles))
-        logger.debug("Filters: %s", self._filters)
+                self._steps.append(
+                    self._make_untrimmed_filter(untrimmed_writer))
+
+            self._steps.append(self._final_filter(outfiles))
+        for i, step in enumerate(self._steps, 1):
+            logger.debug("Pipeline step %d: %s", i, step)
 
     def flush(self) -> None:
         for f in self._textiowrappers:
@@ -291,11 +292,11 @@ class Pipeline(ABC):
         pass
 
     @abstractmethod
-    def _filter_wrapper(self):
+    def _make_filter(self, writer, predicate1: Predicate, predicate2: Predicate):
         pass
 
     @abstractmethod
-    def _untrimmed_filter_wrapper(self):
+    def _make_untrimmed_filter(self, writer):
         pass
 
     @abstractmethod
@@ -306,6 +307,10 @@ class Pipeline(ABC):
     def _create_demultiplexer(self, outfiles: OutputFiles):
         pass
 
+    @abstractmethod
+    def _wrap_single_end_step(self, step: SingleEndStep):
+        pass
+
 
 class SingleEndPipeline(Pipeline):
     """
@@ -332,7 +337,7 @@ class SingleEndPipeline(Pipeline):
             info = ModificationInfo(read)
             for modifier in self._modifiers:
                 read = modifier(read, info)
-            for filter_ in self._filters:
+            for filter_ in self._steps:
                 if filter_(read, info):
                     break
         return (n, total_bp, None)
@@ -348,16 +353,17 @@ class SingleEndPipeline(Pipeline):
         return self.file_opener.dnaio_open_raise_limit(
             file, mode="w", qualities=self.uses_qualities, fileformat="fasta" if force_fasta else None)
 
-    def _filter_wrapper(self):
-        return Redirector
+    def _make_filter(self, writer, predicate1: Predicate, predicate2: Predicate):
+        _ = predicate2
+        return SingleEndFilter(writer, predicate1)
 
-    def _untrimmed_filter_wrapper(self):
-        return Redirector
+    def _make_untrimmed_filter(self, writer):
+        return SingleEndFilter(writer, DiscardUntrimmed())
 
     def _final_filter(self, outfiles: OutputFiles):
         assert outfiles.out2 is None and outfiles.out is not None
         writer = self._open_writer(outfiles.out, force_fasta=outfiles.force_fasta)
-        return NoFilter(writer)
+        return SingleEndSink(writer)
 
     def _create_demultiplexer(self, outfiles: OutputFiles):
         writers = dict()  # type: Dict[Optional[str], Any]
@@ -368,6 +374,9 @@ class SingleEndPipeline(Pipeline):
             writers[name] = self._open_writer(file, force_fasta=outfiles.force_fasta)
         return Demultiplexer(writers)
 
+    def _wrap_single_end_step(self, step: SingleEndStep):
+        return step
+
     @property
     def minimum_length(self):
         return self._minimum_length
@@ -393,7 +402,7 @@ class PairedEndPipeline(Pipeline):
     """
     paired = True
 
-    def __init__(self, pair_filter_mode, file_opener: FileOpener):
+    def __init__(self, file_opener: FileOpener, pair_filter_mode: str):
         super().__init__(file_opener)
         self._modifiers = []  # type: List[PairedEndModifier]
         self._pair_filter_mode = pair_filter_mode
@@ -436,7 +445,7 @@ class PairedEndPipeline(Pipeline):
             info2 = ModificationInfo(read2)
             for modifier in self._modifiers:
                 read1, read2 = modifier(read1, read2, info1, info2)
-            for filter_ in self._filters:
+            for filter_ in self._steps:
                 # Stop writing as soon as one of the filters was successful.
                 if filter_(read1, read2, info1, info2):
                     break
@@ -461,24 +470,26 @@ class PairedEndPipeline(Pipeline):
             interleaved=file2 is None,
         )
 
-    def _filter_wrapper(self, pair_filter_mode=None):
+    def _make_filter(self, writer, predicate1: Predicate, predicate2: Predicate, pair_filter_mode=None):
         if pair_filter_mode is None:
             pair_filter_mode = self._pair_filter_mode
-        return functools.partial(PairedRedirector, pair_filter_mode=pair_filter_mode)
+        return PairedEndFilter(writer, predicate1, predicate2, pair_filter_mode=pair_filter_mode)
 
-    def _untrimmed_filter_wrapper(self):
+    def _make_untrimmed_filter(self, writer):
         """
         Return a different filter wrapper when adapters were given only for R1
         or only for R2 (then override_untrimmed_pair_filter will be set)
         """
-        if self.override_untrimmed_pair_filter:
-            return self._filter_wrapper(pair_filter_mode='both')
-        else:
-            return self._filter_wrapper()
+        return self._make_filter(
+            writer,
+            DiscardUntrimmed(),
+            DiscardUntrimmed(),
+            pair_filter_mode='both' if self.override_untrimmed_pair_filter else None,
+        )
 
     def _final_filter(self, outfiles):
         writer = self._open_writer(outfiles.out, outfiles.out2, force_fasta=outfiles.force_fasta)
-        return PairedNoFilter(writer)
+        return PairedEndSink(writer)
 
     def _create_demultiplexer(self, outfiles):
         def open_writer(file, file2):
@@ -498,6 +509,9 @@ class PairedEndPipeline(Pipeline):
                 writers[name] = open_writer(file, outfiles.demultiplex_out2[name])
             return PairedDemultiplexer(writers)
 
+    def _wrap_single_end_step(self, step: SingleEndStep):
+        return PairedSingleEndStep(step)
+
     @property
     def minimum_length(self):
         return self._minimum_length
@@ -530,7 +544,7 @@ class ReaderProcess(Process):
     and finally sends the stop token -1 ("poison pills") to all connections.
     """
 
-    def __init__(self, path: str, path2: Optional[str], connections, queue, buffer_size, stdin_fd):
+    def __init__(self, path: str, path2: Optional[str], opener: FileOpener, connections, queue, buffer_size, stdin_fd):
         """
         queue -- a Queue of worker indices. A worker writes its own index into this
             queue to notify the reader that it is ready to receive more data.
@@ -543,15 +557,16 @@ class ReaderProcess(Process):
         self.queue = queue
         self.buffer_size = buffer_size
         self.stdin_fd = stdin_fd
+        self._opener = opener
 
     def run(self):
         if self.stdin_fd != -1:
             sys.stdin.close()
             sys.stdin = os.fdopen(self.stdin_fd)
         try:
-            with xopen(self.path, 'rb') as f:
+            with self._opener.xopen(self.path, 'rb') as f:
                 if self.path2:
-                    with xopen(self.path2, 'rb') as f2:
+                    with self._opener.xopen(self.path2, 'rb') as f2:
                         for chunk_index, (chunk1, chunk2) in enumerate(
                                 dnaio.read_paired_chunks(f, f2, self.buffer_size)):
                             self.send_to_worker(chunk_index, chunk1, chunk2)
@@ -630,9 +645,10 @@ class WorkerProcess(Process):
                 self._pipeline.connect_io(infiles, outfiles)
                 (n, bp1, bp2) = self._pipeline.process_reads()
                 self._pipeline.flush()
-                cur_stats = Statistics().collect(n, bp1, bp2, [], self._pipeline._filters)
+                cur_stats = Statistics().collect(n, bp1, bp2, [], self._pipeline._steps)
                 stats += cur_stats
                 self._send_outfiles(outfiles, chunk_index, n)
+                self._pipeline.close()
 
             m = self._pipeline._modifiers
             modifier_stats = Statistics().collect(0, 0, 0 if self._pipeline.paired else None, m, [])
@@ -698,7 +714,7 @@ class PipelineRunner(ABC):
         self._progress = progress
 
     @abstractmethod
-    def run(self):
+    def run(self) -> Statistics:
         pass
 
     @abstractmethod
@@ -740,6 +756,7 @@ class ParallelPipelineRunner(PipelineRunner):
         pipeline: Pipeline,
         infiles: InputPaths,
         outfiles: OutputFiles,
+        opener: FileOpener,
         progress: Progress,
         n_workers: int,
         buffer_size: int = 4 * 1024**2,
@@ -748,8 +765,9 @@ class ParallelPipelineRunner(PipelineRunner):
         self._n_workers = n_workers
         self._need_work_queue = Queue()  # type: Queue
         self._buffer_size = buffer_size
-        self._assign_input(infiles.path1, infiles.path2, infiles.interleaved)
         self._outfiles = outfiles
+        self._opener = opener
+        self._assign_input(infiles.path1, infiles.path2, infiles.interleaved)
 
     def _assign_input(
         self,
@@ -768,7 +786,7 @@ class ParallelPipelineRunner(PipelineRunner):
             # This happens during tests: pytest sets sys.stdin to an object
             # that does not have a file descriptor.
             fileno = -1
-        self._reader_process = ReaderProcess(path1, path2, connw,
+        self._reader_process = ReaderProcess(path1, path2, self._opener, connw,
             self._need_work_queue, self._buffer_size, fileno)
         self._reader_process.daemon = True
         self._reader_process.start()
@@ -798,8 +816,7 @@ class ParallelPipelineRunner(PipelineRunner):
         n = 0  # A running total of the number of processed reads (for progress indicator)
         while connections:
             ready_connections = multiprocessing.connection.wait(connections)
-            for connection in ready_connections:
-                assert isinstance(connection, Connection)
+            for connection in ready_connections:  # type: Any
                 chunk_index = connection.recv()
                 if chunk_index == -1:
                     # the worker is done
@@ -868,7 +885,7 @@ class SerialPipelineRunner(PipelineRunner):
         # TODO
         modifiers = getattr(self._pipeline, "_modifiers", None)
         assert modifiers is not None
-        return Statistics().collect(n, total1_bp, total2_bp, modifiers, self._pipeline._filters)
+        return Statistics().collect(n, total1_bp, total2_bp, modifiers, self._pipeline._steps)
 
     def close(self) -> None:
         self._pipeline.close()
