@@ -5,7 +5,10 @@ from cutadapt.align import (
     PrefixComparer,
     SuffixComparer,
     hamming_sphere,
+    edit_environment,
     SEMIGLOBAL,
+    edit_distance,
+    naive_edit_environment, slow_edit_environment,
 )
 from cutadapt.adapters import Where
 
@@ -36,6 +39,11 @@ class TestAligner:
             Aligner(reference, 0.1, wildcard_ref=True)
         assert "only N wildcards" in info.value.args[0]
 
+    def test_find_empty_in_empty(self):
+        aligner = Aligner("", 0, flags=0, min_overlap=0)
+        result = aligner.locate("")
+        assert (0, 0, 0, 0, 0, 0) == result
+
 
 def test_polya():
     s = 'AAAAAAAAAAAAAAAAA'
@@ -267,3 +275,36 @@ def test_hamming_sphere(sk):
     assert len(result) == 3 ** k * binomial(len(s), k)
     for t in result:
         assert hamming_distance(s, t) == k
+
+
+@pytest.mark.parametrize("k,s", [
+    (0, ""),
+    (0, "A"),
+    (1, "AAA"),
+    (1, "TCATTAGA"),
+    (2, "ACC"),
+    (2, "A" * 10),
+    (3, "TCATTA"),
+])
+@pytest.mark.parametrize("environment_func", [edit_environment, slow_edit_environment])
+def test_edit_environment(k, s, environment_func):
+    result = list(environment_func(s, k))
+    strings, distances, matches = zip(*result)
+    naive = set(naive_edit_environment(s, k))
+    assert len(set(strings)) == len(strings)
+    assert set(strings) == naive
+
+    error_rate = k / len(s) if s else 0.0
+    aligner = Aligner(s, max_error_rate=error_rate, flags=0, min_overlap=len(s))
+    for t, dist, m in result:
+        result = aligner.locate(t)
+        start1, stop1, start2, stop2, matches, errors = result
+        assert errors == dist
+        assert m == matches
+        assert start1 == 0
+        assert stop1 == len(s)
+        assert start2 == 0
+        assert stop2 == len(t)
+        assert edit_distance(s, t) == dist
+        assert m <= len(s), (s, t, dist)
+        assert m <= len(t), (s, t, dist)
