@@ -1,6 +1,9 @@
 import subprocess
 import sys
+import os
 from io import StringIO, BytesIO
+
+import dnaio
 import pytest
 
 from cutadapt.__main__ import main
@@ -46,6 +49,17 @@ def test_quiet_and_report():
     # "Options --quiet and --report cannot be used at the same time"
 
 
+@pytest.mark.parametrize("args", [
+    ("--discard-trimmed", "--discard-untrimmed"),
+    ("--discard-trimmed", "--untrimmed-output", os.devnull),
+    ("--discard-untrimmed", "--untrimmed-output", os.devnull),
+])
+def test_only_one_of_discard_trimmed_discard_untrimmed_untrimmed_output(args):
+    with pytest.raises(SystemExit) as e:
+        main(["-o", os.devnull, *args, datapath("small.fastq")])
+    assert e.value.args[0] == 2
+
+
 def test_debug():
     main(["--debug", "--", datapath("small.fastq")])
 
@@ -81,15 +95,15 @@ def test_lowercase(run):
     run('-a ttagacatatctccgtcg', 'lowercase.fastq', 'small.fastq')
 
 
-def test_rest(run, tmpdir, cores):
+def test_rest(run, tmp_path, cores):
     """-r/--rest-file"""
-    rest = str(tmpdir.join("rest.tmp"))
+    rest = tmp_path / "rest.tmp"
     run(['--cores', str(cores), '-b', 'ADAPTER', '-N', '-r', rest], "rest.fa", "rest.fa")
     assert_files_equal(datapath('rest.txt'), rest)
 
 
-def test_restfront(run, tmpdir):
-    path = str(tmpdir.join("rest.txt"))
+def test_restfront(run, tmp_path):
+    path = tmp_path / "rest.txt"
     run(['-g', 'ADAPTER', '-N', '-r', path], "restfront.fa", "rest.fa")
     assert_files_equal(datapath('restfront.txt'), path)
 
@@ -111,12 +125,13 @@ def test_extensiontxtgz(run):
 
 def test_minimum_length(run):
     """-m/--minimum-length"""
-    run("-m 5 -a TTAGACATATCTCCGTCG", "minlen.fa", "lengths.fa")
+    stats = run("-m 5 -a TTAGACATATCTCCGTCG", "minlen.fa", "lengths.fa")
+    assert stats.written_bp[0] == 45
+    assert stats.written == 6
 
 
-def test_too_short(run, tmpdir, cores):
-    """--too-short-output"""
-    too_short_path = str(tmpdir.join('tooshort.fa'))
+def test_too_short(run, tmp_path, cores):
+    too_short_path = tmp_path / 'tooshort.fa'
     stats = run([
         "--cores", str(cores),
         "-m", "5",
@@ -124,7 +139,19 @@ def test_too_short(run, tmpdir, cores):
         "--too-short-output", too_short_path
     ], "minlen.fa", "lengths.fa")
     assert_files_equal(datapath('tooshort.fa'), too_short_path)
-    assert stats.too_short == 5
+    assert stats.filtered["too_short"] == 5
+
+
+@pytest.mark.parametrize("redirect", (False, True))
+def test_too_short_statistics(redirect):
+    args = ["-a", "TTAGACATATCTCCGTCG", "-m", "24", "-o", os.devnull, datapath("small.fastq")]
+    if redirect:
+        args[:0] = ["--too-short-output", os.devnull]
+    stats = main(args)
+    assert stats.with_adapters[0] == 2
+    assert stats.written == 2
+    assert stats.written_bp[0] == 58
+    assert stats.filtered["too_short"] == 1
 
 
 def test_maximum_length(run):
@@ -132,9 +159,9 @@ def test_maximum_length(run):
     run("-M 5 -a TTAGACATATCTCCGTCG", "maxlen.fa", "lengths.fa")
 
 
-def test_too_long(run, tmpdir, cores):
+def test_too_long(run, tmp_path, cores):
     """--too-long-output"""
-    too_long_path = str(tmpdir.join('toolong.fa'))
+    too_long_path = tmp_path / 'toolong.fa'
     stats = run([
         "--cores", str(cores),
         "-M", "5",
@@ -142,7 +169,7 @@ def test_too_long(run, tmpdir, cores):
         "--too-long-output", too_long_path
     ], "maxlen.fa", "lengths.fa")
     assert_files_equal(datapath('toolong.fa'), too_long_path)
-    assert stats.too_long == 5
+    assert stats.filtered["too_long"] == 5
 
 
 def test_length_tag(run):
@@ -153,19 +180,19 @@ def test_length_tag(run):
 
 
 @pytest.mark.parametrize("length", list(range(3, 11)))
-def test_overlap_a(tmpdir, length):
+def test_overlap_a(tmp_path, length):
     """-O/--overlap with -a"""
     adapter = "catatctccg"
     record = ">read\nGAGACCATTCCAATG" + adapter[:length] + '\n'
-    input = tmpdir.join("overlap.fasta")
-    input.write(record)
+    input = tmp_path / "overlap.fasta"
+    input.write_text(record)
     if length < 7:
         expected = record
     else:
         expected = '>read\nGAGACCATTCCAATG\n'
-    output = tmpdir.join("overlap-trimmed.fasta")
+    output = tmp_path / "overlap-trimmed.fasta"
     main(["-O", "7", "-e", "0", "-a", adapter, "-o", str(output), str(input)])
-    assert expected == output.read()
+    assert expected == output.read_text()
 
 
 def test_overlap_b(run):
@@ -173,6 +200,10 @@ def test_overlap_b(run):
     run("-O 10 -b TTAGACATATCTCCGTCG", "overlapb.fa", "overlapb.fa")
 
 
+def test_trim_n(run):
+    run("--trim-n", "trim-n.fasta", "trim-n.fasta")
+
+
 def test_qualtrim(run):
     """-q with low qualities"""
     run("-q 10 -a XXXXXX", "lowqual.fastq", "lowqual.fastq")
@@ -250,9 +281,9 @@ def test_read_wildcard(run):
     ("-a", "wildcard_adapter.fa"),
     ("-b", "wildcard_adapter_anywhere.fa"),
 ])
-def test_adapter_wildcard(adapter_type, expected, run, tmpdir, cores):
+def test_adapter_wildcard(adapter_type, expected, run, tmp_path, cores):
     """wildcards in adapter"""
-    wildcard_path = str(tmpdir.join("wildcards.txt"))
+    wildcard_path = tmp_path / "wildcards.txt"
     run([
             "--cores", str(cores),
             "--wildcard-file", wildcard_path,
@@ -324,39 +355,52 @@ def test_ellipsis_notation(run):
     run('-a ...TTAGACATAT -g GAGATTGCCA --no-indels', 'no_indels.fasta', 'no_indels.fasta')
 
 
-def test_issue_46(run, tmpdir):
+def test_issue_46(run, tmp_path):
     """issue 46 - IndexError with --wildcard-file"""
     run("--anywhere=AACGTN --wildcard-file={}".format(
-        tmpdir.join("wildcards.txt")), "issue46.fasta", "issue46.fasta")
+        tmp_path / "wildcards.txt"), "issue46.fasta", "issue46.fasta")
 
 
 def test_strip_suffix(run):
     run("--strip-suffix _sequence -a XXXXXXX", "stripped.fasta", "simple.fasta")
 
 
-def test_info_file(run, tmpdir, cores):
+def test_info_file(run, tmp_path, cores):
     # The true adapter sequence in the illumina.fastq.gz data set is
     # GCCTAACTTCTTAGACTGCCTTAAGGACGT (fourth base is different from the sequence shown here)
-    info_path = str(tmpdir.join("info.txt"))
+    info_path = tmp_path / "info.txt"
     run(["--cores", str(cores), "--info-file", info_path, "-a", "adapt=GCCGAACTTCTTAGACTGCCTTAAGGACGT"],
         "illumina.fastq", "illumina.fastq.gz")
-    assert_files_equal(cutpath("illumina.info.txt"), info_path)
+    assert_files_equal(cutpath("illumina.info.txt"), info_path, ignore_trailing_space=True)
 
 
-def test_info_file_times(run, tmpdir, cores):
-    info_path = str(tmpdir.join("info.txt"))
+def test_info_file_times(run, tmp_path, cores):
+    info_path = tmp_path / "info.txt"
     run(["--cores", str(cores), "--info-file", info_path, "--times", "2", "-a", "adapt=GCCGAACTTCTTA",
         "-a", "adapt2=GACTGCCTTAAGGACGT"], "illumina5.fastq", "illumina5.fastq")
-    assert_files_equal(cutpath('illumina5.info.txt'), info_path)
+    assert_files_equal(cutpath('illumina5.info.txt'), info_path, ignore_trailing_space=True)
 
 
-def test_info_file_fasta(run, tmpdir, cores):
-    info_path = str(tmpdir.join("info.txt"))
+def test_info_file_fasta(run, tmp_path, cores):
+    info_path = tmp_path / "info.txt"
     # Just make sure that it runs
     run(["--cores", str(cores), "--info-file", info_path, "-a", "TTAGACATAT", "-g", "GAGATTGCCA", "--no-indels"],
         "no_indels.fasta", "no_indels.fasta")
 
 
+def test_info_file_revcomp(run, tmp_path):
+    info_path = tmp_path / "info-rc.txt"
+    main([
+        "--info-file", str(info_path),
+        "-a", "adapt=GAGTCG",
+        "--revcomp",
+        "--rename={header}",
+        "-o", str(tmp_path / "out.fasta"),
+        datapath("info-rc.fasta")
+    ])
+    assert_files_equal(cutpath("info-rc.txt"), info_path)
+
+
 def test_named_adapter(run):
     run("-a MY_ADAPTER=GCCGAACTTCTTAGACTGCCTTAAGGACGT", "illumina.fastq", "illumina.fastq.gz")
 
@@ -377,9 +421,8 @@ def test_compressed_output(tmp_path, cores, extension):
     main(params)
 
 
-if sys.version_info[:2] >= (3, 3):
-    def test_bzip2_multiblock(run):
-        run('-b TTAGACATATCTCCGTCG', 'small.fastq', 'multiblock.fastq.bz2')
+def test_bzip2_multiblock(run):
+    run("-b TTAGACATATCTCCGTCG", "small.fastq", "multiblock.fastq.bz2")
 
 
 def test_xz(run):
@@ -439,11 +482,14 @@ def test_unconditional_cut_invalid_number():
         main(["-u", "a,b", datapath("small.fastq")])
 
 
-def test_untrimmed_output(run, cores, tmpdir):
-    path = str(tmpdir.join("untrimmed.fastq"))
-    run(["--cores", str(cores), "-a", "TTAGACATATCTCCGTCG", "--untrimmed-output", path],
+def test_untrimmed_output(run, cores, tmp_path):
+    path = tmp_path / "untrimmed.fastq"
+    stats = run(["--cores", str(cores), "-a", "TTAGACATATCTCCGTCG", "--untrimmed-output", path],
         "small.trimmed.fastq", "small.fastq")
     assert_files_equal(cutpath("small.untrimmed.fastq"), path)
+    assert stats.with_adapters[0] == 2
+    assert stats.written == 2
+    assert stats.written_bp[0] == 46
 
 
 def test_adapter_file(run):
@@ -511,11 +557,11 @@ def test_multiple_suffix_adapters_noindels(run):
 
 
 def test_max_n(run):
-    assert run('--max-n 0', 'maxn0.fasta', 'maxn.fasta').too_many_n == 4
-    assert run('--max-n 1', 'maxn1.fasta', 'maxn.fasta').too_many_n == 2
-    assert run('--max-n 2', 'maxn2.fasta', 'maxn.fasta').too_many_n == 1
-    assert run('--max-n 0.2', 'maxn0.2.fasta', 'maxn.fasta').too_many_n == 3
-    assert run('--max-n 0.4', 'maxn0.4.fasta', 'maxn.fasta').too_many_n == 2
+    assert run('--max-n 0', 'maxn0.fasta', 'maxn.fasta').filtered["too_many_n"] == 4
+    assert run('--max-n 1', 'maxn1.fasta', 'maxn.fasta').filtered["too_many_n"] == 2
+    assert run('--max-n 2', 'maxn2.fasta', 'maxn.fasta').filtered["too_many_n"] == 1
+    assert run('--max-n 0.2', 'maxn0.2.fasta', 'maxn.fasta').filtered["too_many_n"] == 3
+    assert run('--max-n 0.4', 'maxn0.4.fasta', 'maxn.fasta').filtered["too_many_n"] == 2
 
 
 def test_quiet_is_quiet():
@@ -528,7 +574,7 @@ def test_quiet_is_quiet():
     try:
         sys.stdout = captured_standard_output
         sys.stderr = captured_standard_error
-        main(['-o', '/dev/null', '--quiet', datapath('small.fastq')])
+        main(['-o', os.devnull, '--quiet', datapath('small.fastq')])
     finally:
         sys.stdout = old_stdout
         sys.stderr = old_stderr
@@ -539,7 +585,7 @@ def test_quiet_is_quiet():
 
 
 def test_x_brace_notation():
-    main(['-o', '/dev/null', '--quiet', '-a', 'X{5}', datapath('small.fastq')])
+    main(['-o', os.devnull, '--quiet', '-a', 'X{5}', datapath('small.fastq')])
 
 
 def test_nextseq(run):
@@ -575,11 +621,11 @@ def test_linked_lowercase(run):
         'linked-lowercase.fasta', 'linked.fasta')
 
 
-def test_linked_info_file(tmpdir):
-    info_path = str(tmpdir.join('info.txt'))
-    main(['-a linkedadapter=^AAAAAAAAAA...TTTTTTTTTT', '--info-file', info_path,
-        '-o', str(tmpdir.join('out.fasta')), datapath('linked.fasta')])
-    assert_files_equal(cutpath('linked-info.txt'), info_path)
+def test_linked_info_file(tmp_path):
+    info_path = tmp_path / 'info.txt'
+    main(['-a linkedadapter=^AAAAAAAAAA...TTTTTTTTTT', '--info-file', str(info_path),
+        '-o', str(tmp_path / 'out.fasta'), datapath('linked.fasta')])
+    assert_files_equal(cutpath('linked-info.txt'), info_path, ignore_trailing_space=True)
 
 
 def test_linked_anywhere():
@@ -614,14 +660,19 @@ def test_negative_length(run):
 
 
 @pytest.mark.timeout(0.5)
-def test_issue_296(tmpdir):
+def test_issue_296(tmp_path):
     # Hang when using both --no-trim and --info-file together
-    info_path = str(tmpdir.join('info.txt'))
-    reads_path = str(tmpdir.join('reads.fasta'))
-    out_path = str(tmpdir.join('out.fasta'))
-    with open(reads_path, 'w') as f:
-        f.write('>read\nCACAAA\n')
-    main(['--info-file', info_path, '--no-trim', '-g', 'TTTCAC', '-o', out_path, reads_path])
+    info_path = tmp_path / 'info.txt'
+    reads_path = tmp_path / 'reads.fasta'
+    out_path = tmp_path / 'out.fasta'
+    reads_path.write_text(">read\nCACAAA\n")
+    main([
+        "--info-file", str(info_path),
+        "--no-trim",
+        "-g", "TTTCAC",
+        "-o", str(out_path),
+        str(reads_path),
+    ])
     # Output should be unchanged because of --no-trim
     assert_files_equal(reads_path, out_path)
 
@@ -636,7 +687,7 @@ def test_adapterx(run):
 
 def test_discard_casava(run):
     stats = run('--discard-casava', 'casava.fastq', 'casava.fastq')
-    assert stats.casava_filtered == 1
+    assert stats.filtered["casava_filtered"] == 1
 
 
 def test_underscore(run):
@@ -649,8 +700,8 @@ def test_cores_autodetect(run):
     run('--cores 0 -b TTAGACATATCTCCGTCG', 'small.fastq', 'underscore_fastq.gz')
 
 
-def test_write_compressed_fastq(cores, tmpdir):
-    main(['--cores', str(cores), '-o', str(tmpdir.join('out.fastq.gz')), datapath('small.fastq')])
+def test_write_compressed_fastq(cores, tmp_path):
+    main(['--cores', str(cores), '-o', str(tmp_path / 'out.fastq.gz'), datapath('small.fastq')])
 
 
 def test_minimal_report(run):
@@ -667,9 +718,9 @@ def test_empty_read_with_wildcard_in_adapter(run):
     run("-g CWC", "empty.fastq", "empty.fastq")
 
 
-def test_print_progress_to_tty(tmpdir, mocker):
+def test_print_progress_to_tty(tmp_path, mocker):
     mocker.patch("cutadapt.utils.sys.stderr").isatty.return_value = True
-    main(["-o", str(tmpdir.join("out.fastq")), datapath("small.fastq")])
+    main(["-o", str(tmp_path / "out.fastq"), datapath("small.fastq")])
 
 
 def test_adapter_order(run):
@@ -677,6 +728,24 @@ def test_adapter_order(run):
     run("-a CCGGG -g ^AAACC", "adapterorder-ag.fasta", "adapterorder.fasta")
 
 
+def test_reverse_complement_no_rc_suffix(run, tmp_path):
+    out_path = tmp_path / "out.fastq"
+    main([
+        "-o", str(out_path),
+        "--revcomp",
+        "--no-index",
+        "--rename", "{header}",
+        "-g", "^TTATTTGTCT",
+        "-g", "^TCCGCACTGG",
+        datapath("revcomp.1.fastq")
+    ])
+    with dnaio.open(out_path) as f:
+        reads = list(f)
+    assert len(reads) == 6
+    assert reads[1].name == "read2/1"
+    assert reads[1].sequence == "ACCATCCGATATGTCTAATGTGGCCTGTTG"
+
+
 def test_reverse_complement_normalized(run):
     stats = run(
         "--revcomp --no-index -g ^TTATTTGTCT -g ^TCCGCACTGG",
@@ -712,18 +781,18 @@ def test_reverse_complement_and_info_file(run, tmp_path, cores):
 
 def test_max_expected_errors(run, cores):
     stats = run("--max-ee=0.9", "maxee.fastq", "maxee.fastq")
-    assert stats.too_many_expected_errors == 2
+    assert stats.filtered["too_many_expected_errors"] == 2
 
 
 def test_max_expected_errors_fasta(tmp_path):
     path = tmp_path / "input.fasta"
     path.write_text(">read\nACGTACGT\n")
-    main(["--max-ee=0.001", "-o", "/dev/null", str(path)])
+    main(["--max-ee=0.001", "-o", os.devnull, str(path)])
 
 
 def test_warn_if_en_dashes_used():
     with pytest.raises(SystemExit):
-        main(["–q", "25", "-o", "/dev/null", "in.fastq"])
+        main(["–q", "25", "-o", os.devnull, "in.fastq"])
 
 
 @pytest.mark.parametrize("opt", ["-y", "--suffix"])
@@ -735,7 +804,7 @@ def test_suffix(opt, run):
 @pytest.mark.parametrize("opt", ["--prefix", "--suffix"])
 def test_rename_cannot_be_combined_with_other_renaming_options(opt):
     with pytest.raises(SystemExit):
-        main([opt, "something", "--rename='{id} {comment} extrainfo'", "-o", "/dev/null", datapath("empty.fastq")])
+        main([opt, "something", "--rename='{id} {comment} extrainfo'", "-o", os.devnull, datapath("empty.fastq")])
 
 
 def test_rename(run):
