@@ -1,3 +1,4 @@
+import sys
 import os.path
 import subprocess
 
@@ -14,9 +15,18 @@ class FilesDifferent(Exception):
     pass
 
 
-def assert_files_equal(path1, path2):
+def assert_files_equal(path1, path2, ignore_trailing_space: bool = False):
+    cmd = ["diff", "-u"]
+    if sys.platform == "win32":
+        cmd.append("--strip-trailing-cr")
+    if ignore_trailing_space:
+        if sys.platform == "darwin":
+            # Ignores too much, but macOS doesn’t have the option below
+            cmd.append("-b")
+        else:
+            cmd.append("--ignore-trailing-space")
     try:
-        subprocess.check_output(['diff', '-u', path1, path2], stderr=subprocess.STDOUT)
+        subprocess.check_output(cmd + [path1, path2], stderr=subprocess.STDOUT)
     except subprocess.CalledProcessError as e:
         raise FilesDifferent('\n' + e.output.decode()) from None
 
